@@ -1,9 +1,12 @@
 import builtins
+import platform
 import sys
 from contextlib import ExitStack
+from functools import partial
+from itertools import chain
 
 from clize import Clize, run
-from clize.errors import MissingValue, UnknownOption
+from clize.errors import ArgumentError, MissingValue, UnknownOption, SetArgumentErrorContext
 from clize.parameters import multi
 from clize.parser import use_mixin, Parameter, NamedParameter
 from pkg_resources import iter_entry_points
@@ -50,7 +53,8 @@ def pretty_syntax_error(source, err):
         print('  %s' % line, file=sys.stderr)
         if err.lineno == lineno and err.offset > 0:
             off = err.offset - 1
-            if err.text is None and sys.version_info < (3, 8):  # pragma: no cover
+            if (err.text is None and sys.version_info < (3, 8)) or \
+                    platform.python_implementation() == 'PyPy':  # pragma: no cover
                 off += 1
             print('  %s^' % (' ' * off), file=sys.stderr)
     print('%s: %s' % (type(err).__name__, err.msg), file=sys.stderr)
@@ -95,9 +99,10 @@ def make_literal(code, env, pipe_name, debuginfo):
 
 
 def make_context():
+    prelude_things = {k: getattr(prelude, k) for k in prelude.__all__}
     context = Context()
     context.update(builtins.__dict__)
-    context.update(prelude.__dict__)
+    context.update(prelude_things)
     return context
 
 
@@ -139,9 +144,10 @@ class _LiteralDecorated(_Decorated):
 class Decorator(NamedParameter):
     LITERAL = False
 
-    def __init__(self, *a, description, decfn, **kw):
+    def __init__(self, *a, description, dec_args, decfn, **kw):
         super().__init__(*a, **kw)
         self.description = description
+        self.dec_args = dec_args
         self.decfn = decfn
         self.marker_class = _Decorated
 
@@ -150,23 +156,56 @@ class Decorator(NamedParameter):
         helper.sections["Decorators"][self.display_name] = (self, '')
         del helper.sections["Options"][self.display_name]
 
+    def format_argument(self, long_alias):
+        if not self.dec_args:
+            return ''
+        return ('=' if long_alias else ' ') + ' '.join(
+            conv.usage_name for conv in self.dec_args)
+
+    def get_all_names(self):
+        names = super().get_all_names()
+        long_alias = any(alias.startswith('--') for alias in self.aliases)  # pragma: no branch
+        return names + self.format_argument(long_alias)
+
     def parse_one_arg(self, ba, arg):
         try:
             if arg[0:2] == '--':
-                return [ba.sig.aliases[arg]]
+                arg, _, val = arg.partition('=')
+                return [ba.sig.aliases[arg]], val
             elif arg[0] == '-':
-                return [ba.sig.aliases['-' + c] for c in arg[1:]]
+                return [ba.sig.aliases['-' + c] for c in arg[1:]], None
             else:
-                return arg
+                return arg, None
         except KeyError as e:
             raise UnknownOption(e.args[0])
 
     def coalesce(self, dec, decseq, funcseq, names):
+        decseq.append(dec)
         funcseq.append(dec.decfn)
         names.append(dec.display_name)
         return dec.marker_class
 
     def read_argument(self, ba, i):
+        with SetArgumentErrorContext(param=self):
+            self._read_argument(ba, i)
+
+    @staticmethod
+    def arg_iterator(start):
+        iterators = [iter(x) for x in start]
+        def iterate():
+            while iterators:
+                while iterators:
+                    try:
+                        v = next(iterators[0])
+                        yield v
+                        break
+                    except StopIteration:
+                        iterators.pop(0)
+        def pushback(x):
+            iterators.insert(0, iter(x))
+        return iterate(), pushback
+
+    def _read_argument(self, ba, i):
         src = None
         io = i
         funcseq = [self.decfn]
@@ -174,42 +213,83 @@ class Decorator(NamedParameter):
         arg = ba.in_args[i]
         decseq = [self]
         cls = self.marker_class
+        stacked = -1
+        eq = -1
+        leader = []
         if arg[1] == '-':
-            i += 1
+            if '=' in arg[2:]:
+                leader.append((i, arg[2:].split('=', 1)[1]))
+                eq = i
             try:
-                arg = ba.in_args[i]
+                i += 1
+                leader.append((i, ba.in_args[i]))
             except LookupError:
                 raise MissingValue
         else:
             if len(arg) >= 3:
-                arg = '-' + arg[2:]
+                leader.append((i, '-' + arg[2:]))
+                stacked = i
             else:
-                i += 1
                 try:
-                    arg = ba.in_args[i]
+                    i += 1
+                    leader.append((i, ba.in_args[i]))
                 except LookupError:
                     raise MissingValue
-        while True:
-            narg = self.parse_one_arg(ba, arg)
-            if isinstance(narg, list):
-                for dec in narg:
-                    if not isinstance(dec, Decorator):
+        ita, pushback = self.arg_iterator([leader, enumerate(ba.in_args[i + 1:], start=i + 1)])
+        for i, arg in ita:
+            if getattr(funcseq[-1], 'dec_args', None):
+                with SetArgumentErrorContext(param=decseq[-1]):
+                    # if this is a reconstructed next option, we must have
+                    # left-composed a short option that takes args, which is
+                    # definitely wrong
+                    if i == stacked:
                         raise MissingValue
-                    cls = decseq[-1].coalesce(dec, decseq, funcseq, names)
+                    ita_ = chain(((i, arg),), ita)
+                    da = self._read_dec_args(ita_, decseq[-1], names)
+                    funcseq[-1] = funcseq[-1](*da)
+                continue
+            if i == eq:
+                raise ArgumentError("{} does not take a value".format(
+                    decseq[-1].display_name))
+            narg, trail = self.parse_one_arg(ba, arg)
+            if trail:
+                pushback([(i, trail)])
+                eq = i
+            if isinstance(narg, list):
+                for ind, dec in enumerate(narg):
+                    with SetArgumentErrorContext(param=decseq[-1]):
+                        if not isinstance(dec, Decorator):
+                            raise MissingValue
+                        if ind > 0 and decseq[-1].dec_args:
+                            raise MissingValue
+                        cls = decseq[-1].coalesce(dec, decseq, funcseq, names)
             elif not decseq[-1].LITERAL:
                 i, src = StepList._read(ba, i)
                 break
             else:
                 src = ba.in_args[i]
                 break
-            i += 1
-            if i >= len(ba.in_args):
-                raise MissingValue
-            arg = ba.in_args[i]
+        else:
+            raise MissingValue
         ba.skip = i - io
         funcseq.reverse()
         ba.args.append(cls(funcseq, src, ' '.join(names)))
 
+    def _read_dec_args(self, ita, dec, names):
+        a = []
+        for elem, converter in zip(ita, dec.dec_args):
+            i, arg = elem
+            name = arg
+            if not name.isdigit():
+                name = repr(name)
+            names.append(name)
+            a.append(converter(arg))
+            if len(a) >= len(dec.dec_args):
+                break
+        else:  # pragma: no cover
+            raise MissingValue
+        return a
+
 
 class LiteralDecorator(Decorator):
     LITERAL = True
@@ -225,9 +305,9 @@ class LiteralDecorator(Decorator):
 def _main(*steps: use_mixin(StepList),
           each_line: 'l' = False,  # noqa: F821
           raw: 'r' = False,  # noqa: F821
-          start: (int, 's') = 0,
-          end: (int, 'e') = None,
-          prelude: (multi(), 'p') = 'pass',
+          start: (int, 's') = 0,  # noqa: F821
+          end: (int, 'e') = None,  # noqa: F821
+          prelude: (multi(), 'p') = 'pass',  # noqa: F821
           pipe_name: Parameter.UNDOCUMENTED = PIPE_NAME,
           no_default_fragments: Parameter.UNDOCUMENTED = False,
           no_exception_handling: Parameter.UNDOCUMENTED = False,
@@ -317,13 +397,15 @@ def _prepare_decorators():
             cls = Decorator
         rv.append(cls(aliases=fn.decorator_names,
                       description=fn.decorator_help,
-                      decfn=fn))
+                      decfn=fn,
+                      dec_args=fn.dec_args))
     return tuple(rv)
 
 
 def _cli():
     return Clize(_main, extra=_prepare_decorators())
 
+
 def main():
     if not spy._dont_load_plugins:
         for entry_point in iter_entry_points('spy.init'):
